import OSS from 'ali-oss';
import { relative } from 'path';
import { obj as through2Obj } from 'through2';
import Ajv from 'ajv';
import mime from 'mime';

import { basicOptionsSchema } from './optionSchema';

const ajv = new Ajv({ allErrors: false });

interface CreateOssOptions {
  accessKeyId: string;
  accessKeySecret: string;
  region: string;
  bucket: string;
}

interface UploadOptions {
  dest: string;
  file: Record<string, any>;
  ossInstance: Record<string, any>;
  ossPutOptions: Record<string, any>;
}

interface BasicOptions{
  prefix: string;
  ossCreateOptions: CreateOssOptions,
  ossPutOptions?: Record<string, any>
}

/**
 * 创建 ali oss 实例
 * @param options 实例配置
 */
export function createOssInstance(options: CreateOssOptions) {
  let aliOssInstance: OSS;
  let error: any;
  return () => {
    if (!aliOssInstance) {
      try {
        aliOssInstance = new OSS({
          accessKeyId: options.accessKeyId,
          accessKeySecret: options.accessKeySecret,
          region: options.region,
          bucket: options.bucket,
        });
        error = null;
      } catch (e) {
        error = e;
      }
    }

    return [aliOssInstance, error];
  };
}

/**
 * oss 上传
 * @param config 上传 put 配置
 */
export async function ossPut(config: UploadOptions) {
  const {
    dest,
    file,
    ossInstance,
    ossPutOptions,
  } = config;
  try {
    await ossInstance.put(dest, file.path, ossPutOptions); // 上传文件
    return [true, ''];
  } catch (e) {
    return [false, e];
  }
}

/**
 * gulp oss 上传插件
 * @param options
 */
export function ossUpload(options: BasicOptions) {
  const isValidOssUploadOptions = ajv.validate(basicOptionsSchema, options);
  if (!isValidOssUploadOptions) {
    const { message = '', dataPath = '' } = ajv.errors?.length ? ajv.errors[0] : {};
    throw new Error(`${dataPath}，${message}`);
  }

  const {
    prefix,
    ossCreateOptions,
    ossPutOptions = {},
  } = options;

  const [ossInstance, ossCreateError] = createOssInstance(ossCreateOptions)();

  if (!ossInstance) throw ossCreateError;

  // 通过through2获取文件
  return through2Obj(async (file, _enc: string, cb) => {
    if (file.isDirectory()) return cb(null, file);
    if (file.isStream()) return cb(new Error('No stream support'));

    // 主动设置资源请响应头。发现如果不主动设置，上传资源类型是 image/png 格式
    const defaultPutOptions = {
      'Content-Type': mime.getType(file.path),
    };

    const [isUploadSuccess, uploadError] = await ossPut({
      dest: `${prefix}/${relative(file.base, file.path).replace(/\\/g, '/')}`,
      file,
      ossInstance,
      ossPutOptions: {
        ...defaultPutOptions,
        ...ossPutOptions,
      },
    });

    if (!isUploadSuccess) return cb(uploadError);

    return cb(null, file);
  });
}
